from __future__ import annotations
import math
import sys
from collections import namedtuple
from dataclasses import dataclass
from typing import List, Optional, Tuple

PixelScale = namedtuple('PixelScale', ['xstep', 'ystep'])

@dataclass
class MapProjection:
    """Records the map projection and the size of the pixels in a layer.

    This superceeeds the old PixelScale class, which will be removed in version 2.0.

    Parameters
    ----------
    name : str
        The map projection used.
    xstep : float
        The number of units horizontal distance a step of one pixel makes in the map projection.
    ystep : float
        The number of units verticle distance a step of one pixel makes in the map projection.

    Attributes
    ----------
    name : str
        The map projection used.
    xstep : float
        The number of units horizontal distance a step of one pixel makes in the map projection.
    ystep : float
        The number of units verticle distance a step of one pixel makes in the map projection.
    """

    name : str
    xstep : float
    ystep : float

    def __eq__(self, other) -> bool:
        if other is None:
            return True
        # to avoid circular dependancies
        from .rounding import are_pixel_scales_equal_enough  # pylint: disable=C0415
        return (self.name == other.name) and \
            are_pixel_scales_equal_enough([self.scale, other.scale])

    @property
    def scale(self) -> PixelScale:
        return PixelScale(self.xstep, self.ystep)

@dataclass
class Area:
    """Class to hold a geospatial area of data in the given projection.

    Parameters
    ----------
    left : float
        Left most point in the projection space
    top : float
        Top most point in the projection space
    right : float
        Right most point in the projection space
    bottom : float
        Bottom most point in the projection space

    Attributes
    ----------
    left : float
        Left most point in the projection space
    top : float
        Top most point in the projection space
    right : float
        Right most point in the projection space
    bottom : float
        Bottom most point in the projection space
    """
    left: float
    top: float
    right: float
    bottom: float

    @staticmethod
    def world() -> Area:
        """Creates an area that covers the entire planet.

        Returns
        -------
        Area
            An area where the extents are nan, but is_world returns true.
        """
        return Area(float("nan"), float("nan"), float("nan"), float("nan"))

    def __hash__(self):
        return (self.left, self.top, self.right, self.bottom).__hash__()

    def __eq__(self, other) -> bool:
        return math.isclose(self.left, other.left, abs_tol=1e-09) and \
            math.isclose(self.right, other.right, abs_tol=1e-09) and \
            math.isclose(self.top, other.top, abs_tol=1e-09) and \
            math.isclose(self.bottom, other.bottom, abs_tol=1e-09)

    def grow(self, offset: float):
        """Expand the area in all directions by the given amount.

        Generates a new area that is an expanded version of the current area.

        Parameters
        ----------
        offset : float
            The amount by which to grow the area

        Returns
        -------
        Area
            A new area of the expanded size.

        """
        return Area(
            left=self.left - offset,
            top=self.top + offset,
            right=self.right + offset,
            bottom=self.bottom - offset
        )

    @property
    def is_world(self) -> bool:
        """Returns true if this is a global area, independant of projection.

        Returns
        -------
        bool
            True is the Area was created with `world` otherwise False.
        """
        return math.isnan(self.left)

    def overlaps(self, other) -> bool:
        """Check if this area overlaps with another area.

        Parameters
        ----------
        other : Area
            The other area to compare this area with

        Returns
        -------
        bool
            True if the two areas intersect, otherwise false.
        """

        if self.is_world or other.is_world:
            return True

        return (
            (self.left <= other.left <= self.right) or
            (self.left <= other.right <= self.right) or
            (other.left <= self.left <= other.right) or
            (other.left <= self.right <= other.right)
        ) and (
            (self.top >= other.top >= self.bottom) or
            (self.top >= other.bottom >= self.bottom) or
            (other.top >= self.top >= other.bottom) or
            (other.top >= self.bottom >= other.bottom)
        )

@dataclass
class Window:
    """Class to hold the pixel dimensions of data in the given projection.

    Parameters
    ----------
    xoff : int
        X axis offset
    yoff : int
        Y axis offset
    xsize : int
        Width of data in pixels
    bottom : float
        Height of data in pixels

    Attributes
    ----------
    xoff : int
        X axis offset
    yoff : int
        Y axis offset
    xsize : int
        Width of data in pixels
    bottom : float
        Height of data in pixels
    """
    xoff: int
    yoff: int
    xsize: int
    ysize: int

    @property
    def as_array_args(self) -> Tuple[int,...]:
        """A tuple containing xoff, yoff, xsize, and ysize."""
        return (self.xoff, self.yoff, self.xsize, self.ysize)

    def __lt__(self, other) -> bool:
        return (self.xsize < other.xsize) and \
            (self.ysize < other.ysize) and \
            (self.xoff >= other.xoff) and \
            (self.yoff >= other.yoff) and \
            ((self.xoff + self.xsize) <= (other.xoff + other.xsize)) and \
            ((self.yoff + self.ysize) <= (other.yoff + other.ysize))

    def __gt__(self, other) -> bool:
        return (self.xsize > other.xsize) and \
            (self.ysize > other.ysize) and \
            (self.xoff <= other.xoff) and \
            (self.yoff <= other.yoff) and \
            ((self.xoff + self.xsize) >= (other.xoff + other.xsize)) and \
            ((self.yoff + self.ysize) >= (other.yoff + other.ysize))

    def __le__(self, other) -> bool:
        return (self.xsize <= other.xsize) and \
            (self.ysize <= other.ysize) and \
            (self.xoff >= other.xoff) and \
            (self.yoff >= other.yoff) and \
            ((self.xoff + self.xsize) <= (other.xoff + other.xsize)) and \
            ((self.yoff + self.ysize) <= (other.yoff + other.ysize))

    def __ge__(self, other) -> bool:
        return (self.xsize >= other.xsize) and \
            (self.ysize >= other.ysize) and \
            (self.xoff <= other.xoff) and \
            (self.yoff <= other.yoff) and \
            ((self.xoff + self.xsize) >= (other.xoff + other.xsize)) and \
            ((self.yoff + self.ysize) >= (other.yoff + other.ysize))

    def grow(self, pixels: int):
        """Expand the area in all directions by the given amount.

        Generates a new window that is an expanded version of the current window.

        Parameters
        ----------
        pixels : int
            The amount by which to grow the window in pixels

        Returns
        -------
        Window
            A new window of the expanded size.

        """
        return Window(
            xoff=self.xoff - pixels,
            yoff=self.xoff - pixels,
            xsize=self.xsize + (2 * pixels),
            ysize=self.ysize + (2 * pixels),
        )

    @staticmethod
    def find_intersection(windows: List) -> "Window":
        if not windows:
            raise ValueError("Expected list of windows")
        # This isn't very pythonic, as it was originally written, but
        # this method gets called a lot (by every Layer.read_array), so not looping
        # over the window list multiple times halves the performance cost (as
        # measured with cProfile).
        left, top, right, bottom = -sys.maxsize, -sys.maxsize, sys.maxsize, sys.maxsize
        for window in windows:
            left = max(left, window.xoff)
            top = max(top, window.yoff)
            right = min(right, window.xoff + window.xsize)
            bottom = min(bottom, window.yoff + window.ysize)
        if (left >= right) or (top >= bottom):
            raise ValueError('No intersection possible')
        return Window(
            left,
            top,
            right - left,
            bottom - top
        )

    @staticmethod
    def find_intersection_no_throw(windows: List) -> Optional["Window"]:
        if not windows:
            raise ValueError("Expected list of windows")
        # This isn't very pythonic, as it was originally written, but
        # this method gets called a lot (by every Layer.read_array), so not looping
        # over the window list multiple times halves the performance cost (as
        # measured with cProfile).
        left, top, right, bottom = -sys.maxsize, -sys.maxsize, sys.maxsize, sys.maxsize
        for window in windows:
            left = max(left, window.xoff)
            top = max(top, window.yoff)
            right = min(right, window.xoff + window.xsize)
            bottom = min(bottom, window.yoff + window.ysize)
        if (left >= right) or (top >= bottom):
            return None
        return Window(
            left,
            top,
            right - left,
            bottom - top
        )



from pathlib import Path

from osgeo import gdal
import tomli as tomllib

try:
    from importlib import metadata
    __version__: str = metadata.version(__name__)
except ModuleNotFoundError:
    pyproject_path = Path(__file__).parent.parent / "pyproject.toml"
    with open(pyproject_path, "rb") as f:
        pyproject_data = tomllib.load(f)
    __version__ = pyproject_data["project"]["version"]

from ._core import read_raster, read_rasters, read_shape, read_shape_like
from .constants import WGS_84_PROJECTION

gdal.UseExceptions()



from pathlib import Path
from typing import Optional, Sequence, Tuple, Union

from .layers.base import YirgacheffeLayer
from .layers.group import GroupLayer, TiledGroupLayer
from .layers.rasters import RasterLayer
from .layers.vectors import VectorLayer
from .window import MapProjection
from .operators import DataType

def read_raster(
    filename: Union[Path,str],
    band: int = 1,
    ignore_nodata: bool = False,
) -> RasterLayer:
    """Open a raster file (e.g., GeoTIFF).

    Parameters
    ----------
    filename : Path
        Path of raster file to open.
    band : int, default=1
        For multi-band rasters, which band to use (defaults to first if not specified)
    ignore_nodata : bool, default=False
        If the GeoTIFF has a NODATA value, don't subsitute that value for NaN

    Returns
    -------
    RasterLayer
        Returns an layer representing the raster data.
    """
    return RasterLayer.layer_from_file(filename, band, ignore_nodata)

def read_rasters(
    filenames : Sequence[Union[Path,str]],
    tiled: bool=False
) -> GroupLayer:
    """Open a set of raster files (e.g., GeoTIFFs) as a single layer.

    Parameters
    ----------
    filenames : List[Path]
        List of paths of raster files to open.
    tiled : bool, default=False
        If you know that the rasters for a regular tileset, then setting this flag allows
        Yirgacheffe to perform certain optimisations that significantly improve performance for
        this use case.

    Returns
    -------
    GroupLayer
        Returns an layer representing the raster data.
    """
    if not tiled:
        return GroupLayer.layer_from_files(filenames)
    else:
        return TiledGroupLayer.layer_from_files(filenames)

def read_shape(
    filename: Union[Path,str],
    projection: Union[Optional[MapProjection],Optional[Tuple[str,Tuple[float,float]]]]=None,
    where_filter: Optional[str] = None,
    datatype: Optional[DataType] = None,
    burn_value: Union[int,float,str] = 1,
) -> VectorLayer:
    """Open a polygon file (e.g., GeoJSON, GPKG, or ESRI Shape File).

    Parameters
    ----------
    filename : Path
        Path of raster file to open.
    projection: MapProjection or tuple, optional
        The map projection to use,
    where_filter : str, optional
        For use with files with many entries (e.g., GPKG), applies this filter to the data.
    datatype: DataType, default=DataType.Byte
        Specify the data type of the raster data generated.
    burn_value: int or float or str, default=1
        The value of each pixel in the polygon.

    Returns
    -------
    VectorLayer
        Returns an layer representing the vector data.
    """

    if projection is not None:
        if not isinstance(projection, MapProjection):
            projection_name, scale_tuple = projection
            projection = MapProjection(projection_name, scale_tuple[0], scale_tuple[1])

    return VectorLayer._future_layer_from_file(
        filename,
        where_filter,
        projection,
        datatype,
        burn_value,
    )

def read_shape_like(
    filename: Union[Path,str],
    like: YirgacheffeLayer,
    where_filter: Optional[str] = None,
    datatype: Optional[DataType] = None,
    burn_value: Union[int,float,str] = 1,
) -> VectorLayer:
    """Open a polygon file (e.g., GeoJSON, GPKG, or ESRI Shape File).

    Parameters
    ----------
    filename : Path
        Path of raster file to open.
    like: YirgacheffeLayer
        Another layer that has a projection and pixel scale set. This layer will
        use the same projection and pixel scale as that one.
    where_filter : str, optional
        For use with files with many entries (e.g., GPKG), applies this filter to the data.
    datatype: DataType, default=DataType.Byte
        Specify the data type of the raster data generated.
    burn_value: int or float or str, default=1
        The value of each pixel in the polygon.

    Returns
    -------
    VectorLayer
        Returns an layer representing the vector data.
    """
    return VectorLayer.layer_from_file_like(
        filename,
        like,
        where_filter,
        datatype,
        burn_value,
    )



from __future__ import annotations
import copy
from pathlib import Path
from typing import Any, List, Optional, Sequence, Union

import numpy as np
from numpy import ma

from ..operators import DataType
from ..rounding import round_down_pixels
from ..window import Area, Window
from .base import YirgacheffeLayer
from .rasters import RasterLayer
from .._backends import backend


class GroupLayerEmpty(ValueError):
    def __init__(self, msg):
        self.msg = msg

class GroupLayer(YirgacheffeLayer):

    @classmethod
    def layer_from_directory(
        cls,
        directory_path: Union[Path,str],
        name: Optional[str] = None,
        matching: str = "*.tif"
    ) -> GroupLayer:
        if directory_path is None:
            raise ValueError("Directory path is None")
        if isinstance(directory_path, str):
            directory_path = Path(directory_path)
        files = list(directory_path.glob(matching))
        if len(files) < 1:
            raise GroupLayerEmpty(directory_path)
        return cls.layer_from_files(files, name)

    @classmethod
    def layer_from_files(
        cls,
        filenames: Sequence[Union[Path,str]],
        name: Optional[str] = None
    ) -> GroupLayer:
        if filenames is None:
            raise ValueError("filenames argument is None")
        rasters: List[YirgacheffeLayer] = [RasterLayer.layer_from_file(x) for x in filenames]
        if len(rasters) < 1:
            raise GroupLayerEmpty("No files found")
        return cls(rasters, name)

    def __init__(
        self,
        layers: List[YirgacheffeLayer],
        name: Optional[str] = None
    ) -> None:
        if not layers:
            raise GroupLayerEmpty("Expected one or more layers")
        if not all(x.map_projection == layers[0].map_projection for x in layers):
            raise ValueError("Not all layers are the same projection/scale")
        for layer in layers:
            if layer._active_area is not None:
                raise ValueError("Layers can not currently be constrained")

        # area/window are superset of all tiles
        union = YirgacheffeLayer.find_union(layers)
        super().__init__(union, layers[0].map_projection, name=name)

        # We store them in reverse order so that from the user's perspective
        # the first layer in the list will be the most important in terms
        # over overlapping.
        self._underlying_layers = copy.copy(layers)
        self._underlying_layers.reverse()
        self.layers = self._underlying_layers

    def _park(self) -> None:
        for layer in self.layers:
            try:
                layer._park()
            except AttributeError:
                pass

    @property
    def datatype(self) -> DataType:
        return DataType.of_gdal(self.layers[0].datatype)

    def set_window_for_intersection(self, new_area: Area) -> None:
        super().set_window_for_intersection(new_area)

        # filter out layers we don't care about
        self.layers = [layer for layer in self._underlying_layers if layer.area.overlaps(new_area)]

    def set_window_for_union(self, new_area: Area) -> None:
        super().set_window_for_union(new_area)

        # filter out layers we don't care about
        self.layers = [layer for layer in self._underlying_layers if layer.area.overlaps(new_area)]

    def reset_window(self) -> None:
        super().reset_window()
        try:
            self.layers = self._underlying_layers
        except AttributeError:
            pass # called from Base constructor before we've added the extra field

    def _read_array_with_window(
        self,
        xoffset: int,
        yoffset: int,
        xsize: int,
        ysize: int,
        window: Window,
    ) -> Any:
        if (xsize <= 0) or (ysize <= 0):
            raise ValueError("Request dimensions must be positive and non-zero")

        scale = self.pixel_scale
        assert scale is not None

        target_window = Window(
            window.xoff + xoffset,
            window.yoff + yoffset,
            xsize,
            ysize
        )

        contributing_layers = []
        for layer in self.layers:
            # Normally this is hidden with set_window_for_...
            adjusted_layer_window = Window(
                layer.window.xoff + \
                    round_down_pixels(((layer.area.left - self._underlying_area.left) / scale.xstep), abs(scale.xstep)),
                layer.window.yoff + \
                    round_down_pixels(((layer.area.top - self._underlying_area.top) / scale.ystep), abs(scale.ystep)),
                layer.window.xsize,
                layer.window.ysize,
            )
            intersection = Window.find_intersection_no_throw([target_window, adjusted_layer_window])
            if intersection is not None:
                contributing_layers.append((layer, adjusted_layer_window, intersection))

        # Adding the numpy arrays over each other is relatively expensive, so if we only intersect
        # with a single layer, turn this into a direct read
        if len(contributing_layers) == 1:
            layer, adjusted_layer_window, intersection = contributing_layers[0]
            if target_window == intersection:
                data = layer._read_array(
                    intersection.xoff - adjusted_layer_window.xoff,
                    intersection.yoff - adjusted_layer_window.yoff,
                    intersection.xsize,
                    intersection.ysize
                )
                if layer.nodata is not None:
                    data[data.isnan()] = 0.0
                return data

        result = np.zeros((ysize, xsize), dtype=float)
        for layer, adjusted_layer_window, intersection in contributing_layers:
            data = layer._read_array(
                intersection.xoff - adjusted_layer_window.xoff,
                intersection.yoff - adjusted_layer_window.yoff,
                intersection.xsize,
                intersection.ysize
            )
            result_x_offset = (intersection.xoff - xoffset) - window.xoff
            result_y_offset = (intersection.yoff - yoffset) - window.yoff
            if layer.nodata is None:
                result[
                    result_y_offset:result_y_offset + intersection.ysize,
                    result_x_offset:result_x_offset + intersection.xsize
                ] = data
            else:
                masked = ma.masked_invalid(data)
                before = result[
                    result_y_offset:result_y_offset + intersection.ysize,
                    result_x_offset:result_x_offset + intersection.xsize
                ]
                merged = ma.where(masked.mask, before, masked)
                result[
                    result_y_offset:result_y_offset + intersection.ysize,
                    result_x_offset:result_x_offset + intersection.xsize
                ] = merged

        return backend.promote(result)

class TileData:
    """This class exists just to let me sort the tiles into the correct order for processing."""

    def __init__(self, data: Any, x: int, y: int):
        self.data = data
        self.x = x
        self.y = y

    @property
    def origin(self):
        return (self.x, self.y)

    @property
    def width(self):
        return self.data.shape[1]

    @property
    def height(self):
        return self.data.shape[0]

    def __cmp__(self, other):
        if self.y != other.y:
            return self.y.cmp(other.y)
        else:
            return self.x.cmp(other.x)

    def __gt__(self, other):
        if self.y > other.y:
            return True
        if self.y < other.y:
            return False
        return self.x.__gt__(other.x)

    def __repr__(self):
        if self.data is not None:
            return f"<Tile: {self.x} {self.y} {self.data.shape[1]} {self.data.shape[0]}>"
        else:
            return "<Tile: sentinal>"

class TiledGroupLayer(GroupLayer):
    """An optimised version of GroupLayer for the case where you have a grid of regular sized
    layers, e.g., map tiles.

    This class exists as assembling arbitrary shaped tiles into a group is quite slow due
    to numpy being slow to modify one array with the contents of another. This class does
    away with that at the expense of needing to know all tiles have the same shape.__abs__()

    Two notes:
    * You can have missing tiles, and it'll fill in zeros.
    * The tiles can overlap - e.g., JRC Annual Change tiles all overlap by a few pixels on all edges.
    """
    def _read_array_with_window(
        self,
        xoffset: int,
        yoffset: int,
        xsize: int,
        ysize: int,
        window: Window,
    ) -> Any:
        if (xsize <= 0) or (ysize <= 0):
            raise ValueError("Request dimensions must be positive and non-zero")

        scale = self.pixel_scale
        assert scale is not None

        target_window = Window(
            window.xoff + xoffset,
            window.yoff + yoffset,
            xsize,
            ysize
        )

        partials: List[TileData] = []
        for layer in self.layers:
            # Normally this is hidden with set_window_for_...
            adjusted_layer_window = Window(
                layer.window.xoff + \
                    round_down_pixels(((layer.area.left - self._underlying_area.left) / scale.xstep), abs(scale.xstep)),
                layer.window.yoff + \
                    round_down_pixels(((layer.area.top - self._underlying_area.top) / scale.ystep), abs(scale.ystep)),
                layer.window.xsize,
                layer.window.ysize,
            )
            intersection = Window.find_intersection_no_throw([target_window, adjusted_layer_window])
            if intersection is None:
                continue
            data = layer._read_array(
                intersection.xoff - adjusted_layer_window.xoff,
                intersection.yoff - adjusted_layer_window.yoff,
                intersection.xsize,
                intersection.ysize
            )
            result_x_offset = (intersection.xoff - xoffset) - window.xoff
            result_y_offset = (intersection.yoff - yoffset) - window.yoff
            partials.append(TileData(data, result_x_offset, result_y_offset))

        sorted_partials = sorted(partials)

        # When tiles overlap (hello JRC Annual Change!), then if the read aligns
        # with the edge of a tile, then we can end up with multiple results at the same
        # offset as we get the dregs of the above/left tile and the bulk of the data from
        # the "obvious" tile. In which case, we should reject the smaller section. If we have
        # two tiles at the same offset and one is not a perfect subset of the other then the
        # tile set we were given is not regularly shaped, and so we should give up.
        combed_partials: List[TileData] = []
        previous_tile = None
        for tile in sorted_partials:
            if previous_tile is None:
                previous_tile = tile
                continue

            if previous_tile.origin == tile.origin:
                if (tile.width >= previous_tile.width) and \
                    (tile.height >= previous_tile.height):
                    previous_tile = tile
                continue

            combed_partials.append(previous_tile)
            previous_tile = tile
        if previous_tile:
            combed_partials.append(previous_tile)

        # Add a terminator to force the last row to be added
        combed_partials.append(TileData(None, 0, -1))
        last_y_offset = None
        last_y_height = 0
        expected_next_x = 0
        expected_next_y = 0
        data = None
        row_chunk: Optional[np.ndarray] = None

        # Allow for reading off top
        if combed_partials:
            if combed_partials[0].y > 0:
                row_chunk = np.zeros((combed_partials[0].y, xsize))
                last_y_offset = 0
                last_y_height = combed_partials[0].y

        for tile in combed_partials:
            if tile.y == last_y_offset:
                assert row_chunk is not None
                if row_chunk.shape[0] < tile.data.shape[0]:
                    assert last_y_height == row_chunk.shape[0]
                    row_chunk = np.vstack(
                        (row_chunk, np.zeros((tile.data.shape[0] - row_chunk.shape[0], row_chunk.shape[1])))
                    )
                    last_y_height = row_chunk.shape[0]
                new_data = tile.data
                if row_chunk.shape[0] != new_data.shape[0]:
                    assert row_chunk.shape[0] > new_data.shape[0]
                    # we have some overlap data from oversized tiles (hello JRC) when there's a GAP in general
                    new_data = np.vstack(
                        (new_data, np.zeros((row_chunk.shape[0] - new_data.shape[0], new_data.shape[1])))
                    )
                assert row_chunk.shape[0] == new_data.shape[0]

                # We're adding a tile to an existing row
                x_offset = expected_next_x - tile.x
                if x_offset == 0:
                    # Tiles line up neatly!
                    row_chunk = np.hstack((row_chunk, new_data))
                    expected_next_x = expected_next_x + new_data.shape[1]
                elif x_offset > 0:
                    # tiles overlap
                    remainder = new_data.shape[1] - xoffset
                    if remainder > 0:
                        subdata = np.delete(new_data, np.s_[0:x_offset], 1)
                        row_chunk = np.hstack((row_chunk, subdata))
                        expected_next_x = expected_next_x + subdata.shape[1]
                else:
                    # Gap between tiles, so fill it before adding new data
                    row_chunk = np.hstack((row_chunk, np.zeros((new_data.shape[0], -x_offset))))
                    row_chunk = np.hstack((row_chunk, new_data))
                    expected_next_x = expected_next_x + new_data.shape[1] + -x_offset
            else:
                # This is a new row, so we need to add the row in progress
                # and start a new one
                if row_chunk is not None:
                    if row_chunk.shape[1] != xsize:
                        assert row_chunk.shape[1] < xsize, f"row is too wide: expected {xsize}, is {row_chunk.shape[1]}"
                        # Missing tile at end of row, so fill in
                        row_chunk = np.hstack((row_chunk, np.zeros((last_y_height, xsize - row_chunk.shape[1]))))
                    if data is None:
                        data = row_chunk
                        expected_next_y += last_y_height
                    else:
                        if last_y_offset == expected_next_y:
                            data = np.vstack((data, row_chunk))
                            expected_next_y += last_y_height
                        else:
                            assert last_y_offset is not None
                            diff = expected_next_y - last_y_offset
                            assert diff > 0, f"{expected_next_y} - {last_y_offset} <= 0 (aka {diff})"
                            subdata = np.delete(row_chunk, np.s_[0:diff], 0)
                            data = np.vstack((data, subdata))
                            expected_next_y += subdata.shape[0]
                if tile.data is not None:
                    if tile.x != 0:
                        row_chunk = np.hstack((np.zeros((tile.data.shape[0], tile.x)), tile.data))
                    else:
                        row_chunk = tile.data
                    last_y_offset = tile.y
                    last_y_height = tile.data.shape[0]
                    expected_next_x = tile.data.shape[1] + tile.x

        assert last_y_offset is not None
        if (last_y_offset + last_y_height) < ysize:
            data = np.vstack((data, np.zeros((ysize - (last_y_offset + last_y_height), xsize))))

        assert data is not None
        assert data.shape == (ysize, xsize)
        return backend.promote(data)



import pytest

# I've no idea why pylint dislikes this particular import but accepts
# other entries in the module.
from yirgacheffe.window import Area # pylint: disable=E0401, E0611

@pytest.mark.parametrize(
    "lhs,rhs,is_equal,overlaps",
    [
        # Obvious equality
        (Area(-10.0, 10.0, 10.0, -10.0), Area(-10.0, 10.0, 10.0, -10.0), True,  True),
        (Area(-9.0, 9.0, 9.0, -9.0),     Area(-10.0, 10.0, 10.0, -10.0), False, True), # subset
        (Area(-9.0, 9.0, -1.0, 1.0),     Area(1.0, -1.0, 9.0, -9.0),     False, False),
        (Area(-10.0, 10.0, 1.0, -10.0),  Area(-1.0, 10.0, 10.0, -10.0),  False, True),
    ]
)
def test_area_operators(lhs: Area, rhs: Area, is_equal: bool, overlaps: bool) -> None:
    assert (lhs == rhs) == is_equal
    assert (lhs != rhs) == (not is_equal)
    assert (lhs.overlaps(rhs)) == overlaps
    assert (rhs.overlaps(lhs)) == overlaps
    assert not lhs.is_world
    assert not rhs.is_world

def test_global_area() -> None:
    area = Area.world()
    assert area.is_world

    other_area = Area(-10.0, 10.0, 10.0, -10.0)
    assert area.overlaps(other_area)
    assert other_area.overlaps(area)




import numpy as np
import yirgacheffe
from yirgacheffe.layers import RasterLayer, ConstantLayer
from yirgacheffe.operators import DataType
from yirgacheffe.window import Area, PixelScale

def test_constant_save() -> None:
    area = Area(left=-1.0, right=1.0, top=1.0, bottom=-1.0)
    scale = PixelScale(0.1, -0.1)
    with RasterLayer.empty_raster_layer(area, scale, DataType.Float32) as result:
        with ConstantLayer(42.0) as c:
            c.save(result)

        expected = np.full((20, 20), 42.0)
        actual = result.read_array(0, 0, 20, 20)

        assert (expected == actual).all()

def test_constant_parallel_save(monkeypatch) -> None:
    area = Area(left=-1.0, right=1.0, top=1.0, bottom=-1.0)
    scale = PixelScale(0.1, -0.1)
    with RasterLayer.empty_raster_layer(area, scale, DataType.Float32) as result:
        with ConstantLayer(42.0) as c:
            with monkeypatch.context() as m:
                m.setattr(yirgacheffe.constants, "YSTEP", 1)
                c.parallel_save(result)

        expected = np.full((20, 20), 42.0)
        actual = result.read_array(0, 0, 20, 20)

        assert (expected == actual).all()



from __future__ import annotations
from math import floor, ceil
from pathlib import Path
from typing import Any, Optional, Union

from skimage import transform
from yirgacheffe.operators import DataType

from ..window import MapProjection, PixelScale, Window
from .rasters import RasterLayer, YirgacheffeLayer
from .._backends import backend


class RescaledRasterLayer(YirgacheffeLayer):
    """RescaledRaster dynamically rescales a raster, so to you to work with multiple layers at
    different scales without having to store unnecessary data. """

    @classmethod
    def layer_from_file(
        cls,
        filename: Union[Path,str],
        pixel_scale: PixelScale,
        band: int = 1,
        nearest_neighbour: bool = True,
    ) -> RescaledRasterLayer:
        src = RasterLayer.layer_from_file(filename, band=band)
        source_projection = src.map_projection
        if source_projection is None:
            raise ValueError("Source raster must have projection and scale")
        target_projection = MapProjection(source_projection.name, pixel_scale.xstep, pixel_scale.ystep)
        return RescaledRasterLayer(src, target_projection, nearest_neighbour, src.name)

    def __init__(
        self,
        src: RasterLayer,
        target_projection: MapProjection,
        nearest_neighbour: bool = True,
        name: Optional[str] = None,
    ):
        super().__init__(
            src.area,
            target_projection,
            name=name
        )

        self._src = src
        self._nearest_neighbour = nearest_neighbour

        src_projection = src.map_projection
        assert src_projection # from raster we should always have one

        self._x_scale = src_projection.xstep / target_projection.xstep
        self._y_scale = src_projection.ystep / target_projection.ystep

    def close(self):
        self._src.close()

    def _park(self):
        self._src._park()

    def _unpark(self):
        self._src._unpark()

    @property
    def datatype(self) -> DataType:
        return self._src.datatype

    def _read_array_with_window(
        self,
        xoffset: int,
        yoffset: int,
        xsize: int,
        ysize: int,
        window: Window,
    ) -> Any:

        # to avoid aliasing issues, we try to scale to the nearest pixel
        # and recrop when scaling bigger

        xoffset = xoffset + window.xoff
        yoffset = yoffset + window.yoff

        src_x_offset = floor(xoffset / self._x_scale)
        src_y_offset = floor(yoffset / self._y_scale)

        diff_x = floor(((xoffset / self._x_scale) - src_x_offset) * self._x_scale)
        diff_y = floor(((yoffset / self._x_scale) - src_y_offset) * self._x_scale)

        src_x_width = ceil((xsize + diff_x) / self._x_scale)
        src_y_width = ceil((ysize + diff_y) / self._y_scale)

        # Get the matching src data
        src_data = backend.demote_array(self._src.read_array(
            src_x_offset,
            src_y_offset,
            src_x_width,
            src_y_width
        ))

        scaled = transform.resize(
            src_data,
            (src_y_width * self._y_scale, src_x_width * self._x_scale),
            order=(0 if self._nearest_neighbour else 1),
            anti_aliasing=(not self._nearest_neighbour)
        )

        return backend.promote(scaled[diff_y:(diff_y + ysize),diff_x:(diff_x + xsize)])



import numpy

from tests.helpers import gdal_dataset_with_data
from yirgacheffe.layers import RasterLayer

def test_add_byte_layers_with_union() -> None:
    data1 = numpy.array([[1, 2, 3, 4,], [5, 6, 7, 8,], [9, 10, 11, 12,], [13, 14, 15, 16,]])
    data2 = numpy.array([[10, 20,], [50, 60,]])

    layer1 = RasterLayer(gdal_dataset_with_data((0.0, 0.0), 1.0, data1))
    layer2 = RasterLayer(gdal_dataset_with_data((1.0, -1.0), 1.0, data2))
    result = RasterLayer.empty_raster_layer_like(layer1)

    layers = [layer1, layer2, result]
    window = RasterLayer.find_union(layers)
    for layer in layers:
        layer.set_window_for_union(window)

    comp = layer1 + layer2
    comp.save(result)

    expected = numpy.array([[1, 2, 3, 4,], [5, 16, 27, 8,], [9, 60, 71, 12,], [13, 14, 15, 16,]])
    result.reset_window()
    actual = result.read_array(0, 0, 4, 4)

    assert (expected == actual).all()

def test_add_byte_layers_with_intersection_with_max_save_raster() -> None:
    data1 = numpy.array([[1, 2, 3, 4,], [5, 6, 7, 8,], [9, 10, 11, 12,], [13, 14, 15, 16,]])
    data2 = numpy.array([[10, 20,], [50, 60,]])

    layer1 = RasterLayer(gdal_dataset_with_data((0.0, 0.0), 1.0, data1))
    layer2 = RasterLayer(gdal_dataset_with_data((1.0, -1.0), 1.0, data2))
    result = RasterLayer.empty_raster_layer_like(layer1)

    layers = [layer1, layer2, result]
    window = RasterLayer.find_intersection(layers)
    for layer in layers:
        layer.set_window_for_intersection(window)

    comp = layer1 + layer2
    comp.save(result)

    expected = numpy.array([[0, 0, 0, 0,], [0, 16, 27, 0,], [0, 60, 71, 0,], [0, 0, 0, 0,]])
    result.reset_window()
    actual = result.read_array(0, 0, 4, 4)

    assert (expected == actual).all()

def test_add_byte_layers_with_intersection_with_min_save_raster() -> None:
    data1 = numpy.array([[1, 2, 3, 4,], [5, 6, 7, 8,], [9, 10, 11, 12,], [13, 14, 15, 16,]])
    data2 = numpy.array([[10, 20,], [50, 60,]])

    layer1 = RasterLayer(gdal_dataset_with_data((0.0, 0.0), 1.0, data1))
    layer2 = RasterLayer(gdal_dataset_with_data((1.0, -1.0), 1.0, data2))
    result = RasterLayer.empty_raster_layer_like(layer2)
    layers = [layer1, layer2, result]

    window = RasterLayer.find_intersection(layers)
    for layer in layers:
        layer.set_window_for_intersection(window)

    comp = layer1 + layer2
    comp.save(result)

    expected = numpy.array([[16, 27,], [60, 71,],])
    result.reset_window()
    actual = result.read_array(0, 0, 2, 2)

    assert (expected == actual).all()



from __future__ import annotations
from math import ceil, floor
from pathlib import Path
from typing import Any, Optional, Tuple, Union
from typing_extensions import NotRequired

from osgeo import gdal, ogr

from ..operators import DataType
from ..window import Area, MapProjection, PixelScale
from .base import YirgacheffeLayer
from .rasters import RasterLayer
from .._backends import backend

def _validate_burn_value(burn_value: Any, layer: ogr.Layer) -> DataType: # pylint: disable=R0911
    if isinstance(burn_value, str):
        # burn value is field name, so validate it
        index = layer.FindFieldIndex(burn_value, True)
        if index < 0:
            raise ValueError("Burn value not found as field")
        # if the user hasn't specified, pick datatype from
        # fiend definition.
        definition = layer.GetLayerDefn()
        field = definition.GetFieldDefn(index)
        typename = field.GetTypeName()
        if typename == "Integer":
            return DataType.Int64
        elif typename == "Real":
            return DataType.Float64
        else:
            raise ValueError(f"Can't set datatype {typename} for burn value {burn_value}")
    elif isinstance(burn_value, int):
        if 0 <= burn_value <= 255:
            return DataType.Byte
        else:
            unsigned = burn_value > 0
            if unsigned:
                if burn_value < (pow(2, 16)):
                    return DataType.UInt16
                elif burn_value < (pow(2, 32)):
                    return DataType.UInt32
                else:
                    return DataType.UInt64
            else:
                if abs(burn_value) < (pow(2, 15)):
                    return DataType.Int16
                elif abs(burn_value) < (pow(2, 31)):
                    return DataType.Int32
                else:
                    return DataType.Int64
    elif isinstance(burn_value, float):
        return DataType.Float64
    else:
        raise ValueError(f"data type of burn value {burn_value} not supported")


class RasteredVectorLayer(RasterLayer):
    """This layer takes a vector file and rasterises it for the given filter. Rasterization
    up front like this is very expensive, so not recommended. Instead you should use
    VectorLayer."""

    @classmethod
    def layer_from_file( # type: ignore[override] # pylint: disable=W0221
        cls,
        filename: Union[Path,str],
        where_filter: Optional[str],
        scale: PixelScale,
        projection: str,
        datatype: Optional[Union[int, DataType]] = None,
        burn_value: Union[int,float,str] = 1,
    ) -> RasteredVectorLayer:
        vectors = ogr.Open(filename)
        if vectors is None:
            raise FileNotFoundError(filename)
        layer = vectors.GetLayer()
        if where_filter is not None:
            layer.SetAttributeFilter(where_filter)

        estimated_datatype = _validate_burn_value(burn_value, layer)
        if datatype is None:
            datatype_arg: DataType = estimated_datatype
        elif isinstance(datatype, int):
            datatype_arg = DataType.of_gdal(datatype)
        else:
            datatype_arg = datatype

        map_projection = MapProjection(projection, scale.xstep, scale.ystep)

        vector_layer = RasteredVectorLayer(
            layer,
            map_projection,
            datatype=datatype_arg,
            burn_value=burn_value
        )

        # this is a gross hack, but unless you hold open the original file, you'll get
        # a SIGSEGV when using the layers from it later, as some SWIG pointers outlive
        # the original object being around
        vector_layer._original = vectors
        return vector_layer

    def __init__(
        self,
        layer: ogr.Layer,
        projection: MapProjection,
        datatype: Union[int, DataType] = DataType.Byte,
        burn_value: Union[int,float,str] = 1,
    ):
        if layer is None:
            raise ValueError('No layer provided')
        self.layer = layer

        self._original: Optional[Any] = None

        if isinstance(datatype, int):
            datatype_arg = DataType.of_gdal(datatype)
        else:
            datatype_arg = datatype

        # work out region for mask
        envelopes = []
        layer.ResetReading()
        feature = layer.GetNextFeature()
        while feature:
            geometry = feature.GetGeometryRef()
            if geometry:
                envelopes.append(geometry.GetEnvelope())
            feature = layer.GetNextFeature()
        if len(envelopes) == 0:
            raise ValueError('No geometry found for')

        # Get the area, but scale it to the pixel resolution that we're using. Note that
        # the pixel scale GDAL uses can have -ve values, but those will mess up the
        # ceil/floor math, so we use absolute versions when trying to round.
        abs_xstep, abs_ystep = abs(projection.xstep), abs(projection.ystep)
        area = Area(
            left=floor(min(x[0] for x in envelopes) / abs_xstep) * abs_xstep,
            top=ceil(max(x[3] for x in envelopes) / abs_ystep) * abs_ystep,
            right=ceil(max(x[1] for x in envelopes) / abs_xstep) * abs_xstep,
            bottom=floor(min(x[2] for x in envelopes) / abs_ystep) * abs_ystep,
        )

        # create new dataset for just that area
        dataset = gdal.GetDriverByName('mem').Create(
            'mem',
            round((area.right - area.left) / abs_xstep),
            round((area.top - area.bottom) / abs_ystep),
            1,
            datatype_arg.to_gdal(),
            []
        )
        if not dataset:
            raise MemoryError('Failed to create memory mask')

        dataset.SetProjection(projection.name)
        dataset.SetGeoTransform([area.left, projection.xstep, 0.0, area.top, 0.0, projection.ystep])
        if isinstance(burn_value, (int, float)):
            gdal.RasterizeLayer(dataset, [1], self.layer, burn_values=[burn_value], options=["ALL_TOUCHED=TRUE"])
        elif isinstance(burn_value, str):
            gdal.RasterizeLayer(dataset, [1], self.layer, options=[f"ATTRIBUTE={burn_value}", "ALL_TOUCHED=TRUE"])
        else:
            raise ValueError("Burn value for layer should be number or field name")
        super().__init__(dataset)


class VectorLayer(YirgacheffeLayer):
    """This layer takes a vector file and rasterises it for the given filter. Rasterization occurs only
    when the data is fetched, so there is no explosive memeory cost, but fetching small units (e.g., one
    line at a time) can be quite slow, so recommended that you fetch reasonable chunks each time (or
    modify this class so that it chunks things internally)."""

    @classmethod
    def layer_from_file_like(
        cls,
        filename: Union[Path,str],
        other_layer: YirgacheffeLayer,
        where_filter: Optional[str]=None,
        datatype: Optional[Union[int, DataType]] = None,
        burn_value: Union[int,float,str] = 1,
    ) -> VectorLayer:
        if other_layer is None:
            raise ValueError("like layer can not be None")
        map_projection = other_layer.map_projection
        if map_projection is None:
            raise ValueError("Reference layer must have projectione")

        vectors = ogr.Open(filename)
        if vectors is None:
            raise FileNotFoundError(filename)
        layer = vectors.GetLayer()
        if where_filter is not None:
            layer.SetAttributeFilter(where_filter)

        if datatype is not None:
            if isinstance(datatype, int):
                datatype = DataType.of_gdal(datatype)

        vector_layer = VectorLayer(
            layer,
            map_projection,
            name=str(filename),
            datatype=datatype if datatype is not None else other_layer.datatype,
            burn_value=burn_value,
            anchor=(other_layer.area.left, other_layer.area.top),
        )

        # this is a gross hack, but unless you hold open the original file, you'll get
        # a SIGSEGV when using the layers from it later, as some SWIG pointers outlive
        # the original object being around
        vector_layer._original = vectors
        vector_layer._dataset_path = filename if isinstance(filename, Path) else Path(filename)
        vector_layer._filter = where_filter
        return vector_layer

    @classmethod
    def layer_from_file(
        cls,
        filename: Union[Path,str],
        where_filter: Optional[str],
        scale: Optional[PixelScale],
        projection: Optional[str],
        datatype: Optional[Union[int, DataType]] = None,
        burn_value: Union[int,float,str] = 1,
        anchor: Tuple[float,float] = (0.0, 0.0)
    ) -> VectorLayer:
        # In 2.0 we need to remove this and migrate to the MapProjection version
        if (projection is None) ^ (scale is None):
            raise ValueError("Either both projection and scale must be provide, or neither")
        if projection is not None and scale is not None:
            map_projection = MapProjection(projection, scale.xstep, scale.ystep)
        else:
            map_projection = None
        return cls._future_layer_from_file(
            filename,
            where_filter,
            map_projection,
            datatype,
            burn_value,
            anchor
        )

    @classmethod
    def _future_layer_from_file(
        cls,
        filename: Union[Path,str],
        where_filter: Optional[str],
        projection: Optional[MapProjection],
        datatype: Optional[Union[int, DataType]] = None,
        burn_value: Union[int,float,str] = 1,
        anchor: Tuple[float,float] = (0.0, 0.0)
    ) -> VectorLayer:
        try:
            vectors = ogr.Open(filename)
        except RuntimeError as exc:
            # With exceptions on GDAL now returns the wrong (IMHO) exception
            raise FileNotFoundError(filename) from exc
        layer = vectors.GetLayer()
        if where_filter is not None:
            layer.SetAttributeFilter(where_filter)

        estimated_datatype = _validate_burn_value(burn_value, layer)
        if datatype is None:
            datatype = estimated_datatype

        if isinstance(datatype, int):
            datatype_arg = DataType.of_gdal(datatype)
        else:
            datatype_arg = datatype

        vector_layer = VectorLayer(
            layer,
            projection,
            name=str(filename),
            datatype=datatype_arg,
            burn_value=burn_value,
            anchor=anchor
        )

        # this is a gross hack, but unless you hold open the original file, you'll get
        # a SIGSEGV when using the layers from it later, as some SWIG pointers outlive
        # the original object being around
        vector_layer._original = vectors
        vector_layer._dataset_path = filename if isinstance(filename, Path) else Path(filename)
        vector_layer._filter = where_filter
        return vector_layer

    def __init__(
        self,
        layer: ogr.Layer,
        projection: Optional[MapProjection],
        name: Optional[str] = None,
        datatype: Union[int,DataType] = DataType.Byte,
        burn_value: Union[int,float,str] = 1,
        anchor: Tuple[float,float] = (0.0, 0.0)
    ):
        if layer is None:
            raise ValueError('No layer provided')
        self.layer = layer
        self.name = name

        if isinstance(datatype, int):
            self._datatype = DataType.of_gdal(datatype)
        else:
            self._datatype = datatype

        # If the burn value is a number, use it directly, if it's a string
        # then assume it is a column name in the dataset
        self.burn_value = burn_value

        self._original = None
        self._dataset_path: Optional[Path] = None
        self._filter: Optional[str] = None
        self._anchor: Tuple[float,float] = (0.0, 0.0)

        # work out region for mask
        envelopes = []
        layer.ResetReading()
        feature = layer.GetNextFeature()
        while feature:
            geometry = feature.GetGeometryRef()
            if geometry:
                envelopes.append(geometry.GetEnvelope())
            feature = layer.GetNextFeature()
        if len(envelopes) == 0:
            raise ValueError('No geometry found')
        self._anchor = anchor
        self._envelopes = envelopes

        # if projection is not None:
        #     # Get the area, but scale it to the pixel resolution that we're using. Note that
        #     # the pixel scale GDAL uses can have -ve values, but those will mess up the
        #     # ceil/floor math, so we use absolute versions when trying to round.
        #     abs_xstep, abs_ystep = abs(projection.xstep), abs(projection.ystep)
#
        #     # Lacking any other reference, we will make the raster align with
        #     # (0.0, 0.0), if sometimes we want to align with an existing raster, so if
        #     # an anchor is specified, ensure we use that as our pixel space alignment
        #     x_anchor = anchor[0]
        #     y_anchor = anchor[1]
        #     left_shift = x_anchor - abs_xstep
        #     right_shift = x_anchor
        #     top_shift = y_anchor
        #     bottom_shift = y_anchor - abs_ystep
#
        #     area = Area(
        #         left=(floor((min(x[0] for x in envelopes) - left_shift) / abs_xstep) * abs_xstep) + left_shift,
        #         top=(ceil((max(x[3] for x in envelopes) - top_shift) / abs_ystep) * abs_ystep) + top_shift,
        #         right=(ceil((max(x[1] for x in envelopes) - right_shift) / abs_xstep) * abs_xstep) + right_shift,
        #         bottom=(floor((min(x[2] for x in envelopes) - bottom_shift) / abs_ystep) * abs_ystep) + bottom_shift,
        #     )
        # else:
            # If we don't have  a projection just go with the idealised area
        area = Area(
            left=floor(min(x[0] for x in envelopes)),
            top=ceil(max(x[3] for x in envelopes)),
            right=ceil(max(x[1] for x in envelopes)),
            bottom=floor(min(x[2] for x in envelopes)),
        )

        super().__init__(area, projection)


    def _get_operation_area(self, projection: Optional[MapProjection]=None) -> Area:
        if self._projection is not None and projection is not None and self._projection != projection:
            raise ValueError("Calculation projection does not match layer projection")

        target_projection = projection if projection is not None else self._projection

        if target_projection is None:
            if self._active_area is not None:
                return self._active_area
            else:
                return self._underlying_area
        else:
            # Get the area, but scale it to the pixel resolution that we're using. Note that
            # the pixel scale GDAL uses can have -ve values, but those will mess up the
            # ceil/floor math, so we use absolute versions when trying to round.
            abs_xstep, abs_ystep = abs(target_projection.xstep), abs(target_projection.ystep)

            # Lacking any other reference, we will make the raster align with
            # (0.0, 0.0), if sometimes we want to align with an existing raster, so if
            # an anchor is specified, ensure we use that as our pixel space alignment
            x_anchor = self._anchor[0]
            y_anchor = self._anchor[1]
            left_shift = x_anchor - abs_xstep
            right_shift = x_anchor
            top_shift = y_anchor
            bottom_shift = y_anchor - abs_ystep

            envelopes = self._envelopes
            return Area(
                left=(floor((min(x[0] for x in envelopes) - left_shift) / abs_xstep) * abs_xstep) + left_shift,
                top=(ceil((max(x[3] for x in envelopes) - top_shift) / abs_ystep) * abs_ystep) + top_shift,
                right=(ceil((max(x[1] for x in envelopes) - right_shift) / abs_xstep) * abs_xstep) + right_shift,
                bottom=(floor((min(x[2] for x in envelopes) - bottom_shift) / abs_ystep) * abs_ystep) + bottom_shift,
            )

    @property
    def area(self) -> Area:
        if self._active_area is not None:
            return self._active_area
        else:
            return self._get_operation_area()

    def __getstate__(self) -> object:
        # Only support pickling on file backed layers (ideally read only ones...)
        if self._dataset_path is None or not self._dataset_path.exists():
            raise ValueError("Can not pickle layer that is not file backed.")
        odict = self.__dict__.copy()
        del odict['_original']
        del odict['layer']
        return odict

    def __setstate__(self, state):
        vectors = ogr.Open(state['_dataset_path'])
        if vectors is None:
            raise FileNotFoundError(f"Failed to open pickled vectors {state['_dataset_path']}")
        self.__dict__.update(state)
        self._original = vectors
        self.layer = vectors.GetLayer()
        if self._filter is not None:
            self.layer.SetAttributeFilter(self._filter)

    def _park(self):
        self._original = None

    def _unpark(self):
        if getattr(self, "_original", None) is None:
            try:
                self._original = ogr.Open(self._dataset_path)
            except RuntimeError as exc:
                raise FileNotFoundError(f"Failed to open pickled layer {self._dataset_path}") from exc
            self.layer = self._original.GetLayer()
            if self._filter is not None:
                self.layer.SetAttributeFilter(self._filter)

    @property
    def datatype(self) -> DataType:
        return self._datatype

    def _read_array_for_area(
        self,
        target_area: Area,
        target_projection: Optional[MapProjection],
        x: int,
        y: int,
        width: int,
        height: int,
    ) -> Any:
        projection = target_projection if target_projection is not None else self._projection
        assert projection is not None

        if self._original is None:
            self._unpark()
        if (width <= 0) or (height <= 0):
            raise ValueError("Request dimensions must be positive and non-zero")

        # I did try recycling this object to save allocation/dealloction, but in practice it
        # seemed to only make things slower (particularly as you need to zero the memory each time yourself)
        dataset = gdal.GetDriverByName('mem').Create(
            'mem',
            width,
            height,
            1,
            self.datatype.to_gdal(),
            []
        )
        if not dataset:
            raise MemoryError('Failed to create memory mask')

        dataset.SetProjection(projection.name)
        dataset.SetGeoTransform([
            target_area.left + (x * projection.xstep),
            projection.xstep,
            0.0,
            target_area.top + (y * projection.ystep),
            0.0,
            projection.ystep
        ])
        if isinstance(self.burn_value, (int, float)):
            gdal.RasterizeLayer(dataset, [1], self.layer, burn_values=[self.burn_value], options=["ALL_TOUCHED=TRUE"])
        elif isinstance(self.burn_value, str):
            gdal.RasterizeLayer(dataset, [1], self.layer, options=[f"ATTRIBUTE={self.burn_value}", "ALL_TOUCHED=TRUE"])
        else:
            raise ValueError("Burn value for layer should be number or field name")

        res = backend.promote(dataset.ReadAsArray(0, 0, width, height))
        return res

    def _read_array_with_window(self, _x, _y, _width, _height, _window) -> Any:
        assert NotRequired

    def _read_array(self, x: int, y: int, width: int, height: int) -> Any:
        return self._read_array_for_area(self.area, None, x, y, width, height)

    def read_array(self, x: int, y: int, width: int, height: int) -> Any:
        res = self._read_array(x, y, width, height)
        return backend.demote_array(res)



import os

BACKEND = os.environ.get("YIRGACHEFFE_BACKEND", "NUMPY").upper()

match BACKEND:
    case "MLX":
        from . import mlx
        backend = mlx
    case "NUMPY":
        from . import numpy
        backend = numpy # type: ignore[misc]
    case _:
        raise NotImplementedError("Only NUMPY and MLX backends supported")



from osgeo import ogr

from ..window import PixelScale
from .base import YirgacheffeLayer
from .rasters import RasterLayer, InvalidRasterBand
from .rescaled import RescaledRasterLayer
from .vectors import RasteredVectorLayer, VectorLayer
from .area import UniformAreaLayer
from .constant import ConstantLayer
from .group import GroupLayer, TiledGroupLayer
try:
    from .h3layer import H3CellLayer
except ModuleNotFoundError:
    pass
